package demux.microsoft;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import ui.GUIDToHex;

import muxer.MuxFilter;
import muxer.MuxManager;

import common.AudioStreamDescriptor;
import common.AvcVideoAnalyzer;
import common.JobStreamI;
import common.StreamDescriptor;
import common.StreamIdentifier;
import common.Util;
import common.VideoStreamDescriptor;
import common.StreamIdentifier.AdvancedAudioCodecAnalyzer;
import common.StreamIdentifier.DolbyDigitalAnalzyer;
import common.StreamIdentifier.MpegAudioAnalyzer;
import common.StreamIdentifier.MpegVideoAnalyzer;
import common.StreamIdentifier.StreamAnalyzerI;
import common.StreamIdentifier.VC1VideoAnalyzer;
import common.mpeg.MpegConstants;
import common.mpeg.MpegUtil;
import common.packet.Packet;
import common.packet.PacketPart;
import common.packet.PesPacket;

import demux.Demuxer;
import demux.pes.PesUtil;

public class WtvToPesDemuxer extends Demuxer {
	public static final byte[] EOS_GUID=new byte[16];
	public static final byte[] DATA_GUID=Util.hexToBytes("95 C3 D2 C2 7E 9A DA 11 8B F7 00 07 E9 5E AD 8D");
	public static final byte[] TIMING_GUID=Util.hexToBytes("5B 05 E6 1B 97 A9 49 43 88 17 1A 65 5A 29 8A 97"); 	
	public static final byte[] STREAM_INFO_GUID=Util.hexToBytes("ED A4 13 23 2D BF 4F 45 AD 8A D9 5B A7 F9 1F EE");
	public static final byte[] STREAM_INFO_ALL_GUID=Util.hexToBytes("A2 C3 D2 C2 7E 9A DA 11 8B F7 00 07 E9 5E AD 8D");
	public static final byte[] LANGUAGE_GUID=Util.hexToBytes("6D 66 92 E2 02 9C 8D 44 AA 8D 78 1A 93 FD C3 95");
	
	public static final byte[] EXTEND_MEDIA_INFO=Util.hexToBytes("6F B3 39 67 5F 1D C2 4A 81 92 28 BB 0E 73 D1 6A");
	public static final byte[] EXTEND_AUDIO_MARKER=Util.hexToBytes("28 BD AD 46 D0 6F 96 47 93 B2 15 5C 51 DC 04 8D");//Follows Audio Stream
	//public static final byte[] STREAM_VIDEO=Util.hexToBytes("76 69 64 73");
	//public static final byte[] STREAM_AUDIO=Util.hexToBytes("61 75 64 73");
																															
	//Media Types
	public static final byte[] MEDIATYPE_Audio			=Util.hexToBytes("61 75 64 73 00 00 10 00 80 00 00 AA 00 38 9B 71");//73647561-0000-0010-8000-00AA00389B71
	public static final byte[] MEDIATYPE_Video			=Util.hexToBytes("76 69 64 73 00 00 10 00 80 00 00 AA 00 38 9B 71");//73646976-0000-0010-8000-00AA00389B71
	public static final byte[] MEDIATYPE_MSTVCaption	=Util.hexToBytes("89 8A 8B B8 49 B0 80 4C AD CF 58 98 98 5E 22 C1");//b88b8a89-b049-4c80-adcf-5898985e22c1
	public static final byte[] MEDIATYPE_MPEG2_SectionsAndTables	=Util.hexToBytes("6C 17 5F 45 06 4B CE 47 9A EF 8C AE F7 3D F7 B5");//455f176c-4b06-47ce-9aef-8caef73df7b5
	public static final byte[] MEDIATYPE_MPEG2_PES		=Util.hexToBytes("20 80 6D E0 46 DB CF 11 B4 D1 00 80 5F 6C BB EA");//e06d8020-db46-11cf-b4d1-00805f6cbbea
	//Audio Subtypes
	public static final byte[] MEDIASUBTYPE_DOLBY_AC3	=Util.hexToBytes("2C 80 6D E0 46 DB CF 11 B4 D1 00 80 5F 6C BB EA");//e06d802c-db46-11cf-b4d1-00805f6cbbea
	public static final byte[] MEDIASUBTYPE_DOLBY_DDPLUS=Util.hexToBytes("AF 87 FB A7 02 2D FB 42 A4 D4 05 CD 93 84 3B DD");//a7fb87af-2d02-42fb-a4d4-05cd93843bdd
	public static final byte[] MEDIASUBTYPE_MPEG1Packet	=Util.hexToBytes("81 EB 36 E4 4F 52 CE 11 9F 53 00 20 AF 0B A7 70");//e436eb80-524f-11ce-9f53-0020af0ba770
	public static final byte[] MEDIASUBTYPE_MPEG2_AUDIO =Util.hexToBytes("2B 80 6D E0 46 DB CF 11 B4 D1 00 80 5F 6C BB EA");//e06d802b-db46-11cf-b4d1-00805f6cbbea
	public static final byte[] MEDIASUBTYPE_LATM_AAC_2	=Util.hexToBytes("02 16 00 00 00 00 10 00 80 00 00 AA 00 38 9B 71");//00001602-0000-0010-8000-00aa00389b71
	//Video Subtypes
	public static final byte[] MEDIASUBTYPE_H264		=Util.hexToBytes("48 32 36 34 00 00 10 00 80 00 00 AA 00 38 9B 71");//34363248-0000-0010-8000-00aa00389b71
	public static final byte[] MEDIASUBTYPE_MPEG2_VIDEO	=Util.hexToBytes("26 80 6D E0 46 DB CF 11 B4 D1 00 80 5F 6C BB EA");//e06d8026-db46-11cf-b4d1-00805f6cbbea
	
	public static final String LANGUAGE_UNKNOWN="Unknown";
	
	private static final int MAX_STREAMS=0x80;
	private static final int STREAM_MASK=MAX_STREAMS-1;
	
	private StreamDescriptor[] avStreams=new StreamDescriptor[MAX_STREAMS];
	
	public void parse(JobStreamI stream, MuxManager muxManager) throws IOException {
		Map<Integer, StreamDescriptor> muxedStreamDescriptorMap=muxManager.getMuxedStreamDescriptorMap();
		PesPacket[] streamPackets=new PesPacket[MAX_STREAMS];
		boolean[] usedStreams=null;
		if (muxedStreamDescriptorMap!=null) {
			avStreams=new StreamDescriptor[MAX_STREAMS];
			for (StreamDescriptor sd : muxedStreamDescriptorMap.values()) {
				//A little confusing.  Since the avStreams pull double duty language and selected streams, 
				//we need to make sure that all selected streams have a language on the second pass
				avStreams[sd.getStreamId()]=sd;
			}
		} else {
			usedStreams=new boolean[MAX_STREAMS];
		}
		int index=0;
		WtvFileStream wtvStream=(WtvFileStream)stream;
		int rc=MuxManager.CONTINUE;
		stream.open();
		//stream.skip(0x40000); Handled by WtvFileStream
		byte guid[]=new byte[16];
		
		while (!stream.isEOS() && rc!=MuxManager.ABORT) {
			/*ByteBuffer header=stream.slice(32);
			header.order(ByteOrder.LITTLE_ENDIAN);
			header.get(guid);*/
			ByteBuffer buffer=wtvStream.getSizedBuffer(32);
			buffer.order(ByteOrder.LITTLE_ENDIAN);
			buffer.get(guid); 
			/*System.out.print(Long.toString(((WtvFileStream)stream).getFilePos(),16) + "-");
			Util.printHex(guid, 0, guid.length);
			System.out.println();*/
			int len=buffer.getInt() - 32;
			int pad=len%8==0 ? 0 : 8- (len % 8);
			int streamId=buffer.getInt() & STREAM_MASK;
			buffer.getLong(); //AdvanceBuffer 8 bytes for 32 = 16 + 4 + 4 + 8;
			
			if (isCopied(streamId) && avStreams[streamId] != null && Arrays.equals(guid, TIMING_GUID)) {
				stream.skip(8); //Unknown
				len -=8;
				//ByteBuffer timeChunk=stream.slice(len);
				//timeChunk.order(ByteOrder.LITTLE_ENDIAN);
				buffer=wtvStream.getSizedBuffer(len + 32);
				buffer.order(ByteOrder.LITTLE_ENDIAN);
				long time=buffer.getLong(); // nano x 10;
				buffer.position(buffer.position()+len - 8);
				//ByteBuffer header=stream.slice(32);
				buffer.get(guid);
				len=buffer.getInt() - 32;
				pad=len%8==0 ? 0 : 8- (len % 8);
				if (Arrays.equals(guid, DATA_GUID)) {
					PesPacket pesPacket=streamPackets[streamId];
					//Should do sanity check
					streamId=buffer.getInt() & STREAM_MASK;
					buffer.getLong(); //AdvanceBuffer 8 bytes for 32 = 16 + 4 + 4 + 8;
					//Zero length packets exist!
					if (len > 0) {
						if (time == -1) {
							if (pesPacket != null) {
								pesPacket.getPayload().appendFrom(stream.slice(len));
								len=0;
							}
						} else {
							if (pesPacket != null) {
								muxManager.addPacket(pesPacket);
							}
							//Create new packet
							pesPacket=new PesPacket(streamId, index++);
							/*if (logger.isTrace()) {
								logger.trace("0x" + Integer.toString(streamId, 16) + ": " + Util.timeToString(time/10000));
							}*/
							pesPacket.setHeader(PesUtil.getPesHeader(avStreams[streamId], MpegUtil.timeToPesTime(time/10000)));
							pesPacket.setPayload(new PacketPart(stream.slice(len)));
							streamPackets[streamId]=pesPacket;
							len=0;
						}
					}

				} else {
					throw new RuntimeException("Expected Data after timing chunk");
				}
			} else if (Arrays.equals(guid, EOS_GUID)) {
				break;
			} else if (muxedStreamDescriptorMap==null) {
				if (Arrays.equals(guid, STREAM_INFO_GUID)) {
					//This seems to indicate the streams that are actually used
					usedStreams[streamId]=true;
				} else if (Arrays.equals(guid, STREAM_INFO_ALL_GUID) && usedStreams[streamId]) {
					//This seems to indicate the streams that are actually used
					stream.skip(0xc);
					buffer=stream.slice(len-0xc);
					buffer.order(ByteOrder.LITTLE_ENDIAN);
					len=0;
					byte streamType[]=new byte[16];
					buffer.get(streamType);
					if (Arrays.equals(streamType, MEDIATYPE_Audio)) {
						avStreams[streamId]=new AudioStreamDescriptor(streamId);
						buffer.get(streamType);
						if (Arrays.equals(streamType, EXTEND_AUDIO_MARKER)) {
							buffer.position(buffer.position()+0xc);
							buffer.get(streamType);
							if (!Arrays.equals(streamType, EXTEND_MEDIA_INFO)) {
								logger.debug("Expected EXTEND_MEDIA_INFO");
							}
							int infoLen=buffer.getInt();
							//Skip the WaveFormatEx and get the media GUID
							int skipBytes=infoLen-0x20;
							buffer.position(buffer.position()+skipBytes);
							buffer.get(streamType);
						}
						if (Arrays.equals(streamType, MEDIASUBTYPE_DOLBY_AC3)) {
							avStreams[streamId].setCodec(StreamIdentifier.DolbyDigitalAnalzyer.AC3_CODEC_NAME);
							avStreams[streamId].setTsType(MpegConstants.TS_STREAM_TYPE_AC3_AUDIO);
						} else if (Arrays.equals(streamType, MEDIASUBTYPE_DOLBY_DDPLUS)) {
							avStreams[streamId].setCodec(StreamIdentifier.DolbyDigitalAnalzyer.EAC3_CODEC_NAME);
							avStreams[streamId].setTsType(MpegConstants.TS_STREAM_TYPE_EAC3_AUDIO);
						} else if (Arrays.equals(streamType, MEDIASUBTYPE_MPEG1Packet)) {
							avStreams[streamId].setCodec(StreamIdentifier.MpegAudioAnalyzer.CODEC_NAME);
							avStreams[streamId].setTsType(MpegConstants.TS_STREAM_TYPE_MPEG1_AUDIO);
						} else if (Arrays.equals(streamType, MEDIASUBTYPE_MPEG2_AUDIO)) {
							avStreams[streamId].setCodec(StreamIdentifier.MpegAudioAnalyzer.CODEC_NAME);
							avStreams[streamId].setTsType(MpegConstants.TS_STREAM_TYPE_MPEG2_AUDIO);
						} else if (Arrays.equals(streamType, MEDIASUBTYPE_LATM_AAC_2)) {
							avStreams[streamId].setCodec(StreamIdentifier.AdvancedAudioCodecAnalyzer.CODEC_NAME);
							avStreams[streamId].setTsType(MpegConstants.TS_STREAM_TYPE_LATM_AUDIO);
						} else {
							logger.info("Unknown Audio type: " + GUIDToHex.toGuid(streamType));
						}
					} else if (Arrays.equals(streamType, MEDIATYPE_Video)) {
						avStreams[streamId]=new VideoStreamDescriptor(streamId);
					} else if (Arrays.equals(streamType, MEDIATYPE_MSTVCaption)) {
						avStreams[streamId]=new StreamDescriptor(streamId);
						avStreams[streamId].setCodec("MS TV Caption");
					} else if (Arrays.equals(streamType, MEDIATYPE_MPEG2_SectionsAndTables)) {
						avStreams[streamId]=new StreamDescriptor(streamId);
						avStreams[streamId].setCodec("MPEG2 Sections and Tables");
					} else if (Arrays.equals(streamType, MEDIATYPE_MPEG2_PES)) {
						avStreams[streamId]=new StreamDescriptor(streamId);
						avStreams[streamId].setCodec("MPEG2 PES");
					} else {
						avStreams[streamId]=new StreamDescriptor(streamId);
						logger.info("Unknown Stream Type: " + GUIDToHex.toGuid(streamType));
					}
					
				} else if (Arrays.equals(guid, LANGUAGE_GUID)) {
					buffer=wtvStream.getSizedBuffer(16);
					len -=16;
					buffer.getLong(); //Unknown
					short langCode=buffer.getShort(); //Windows Language Code
					buffer.getShort(); //Unknown
					StringBuffer sb=new StringBuffer(4);
					for (int i=0; i<4; i++) {
						char c=(char)buffer.get();
						if (c > 0) {
							sb.append(c);
						}
					}
					if (sb.length() > 0) {
						avStreams[streamId].setLanguage(sb.toString());
					}
				}
			} 
			stream.skip(len + pad);
		}
		for (int i=0;i<streamPackets.length;i++) {
			if (streamPackets[i] != null) {
				muxManager.addPacket(streamPackets[i]);
			}
		}
		muxManager.streamEnded();
		
	}
	private class WtvStreamIdentifier extends MuxFilter {
		HashMap<Byte, StreamAnalyzerI> audioAnalyzerMap=new HashMap<Byte, StreamAnalyzerI>();
		
		private WtvStreamIdentifier() {
			super();
			MpegAudioAnalyzer mpegAudioAnalyzer=new MpegAudioAnalyzer();
			mpegAudioAnalyzer.setSyncOffsets(new int[]{0,4});
			audioAnalyzerMap.put(MpegConstants.TS_STREAM_TYPE_MPEG1_AUDIO, mpegAudioAnalyzer);
			audioAnalyzerMap.put(MpegConstants.TS_STREAM_TYPE_MPEG2_AUDIO, mpegAudioAnalyzer);
			audioAnalyzerMap.put(MpegConstants.TS_STREAM_TYPE_AAC_AUDIO, new AdvancedAudioCodecAnalyzer());
			audioAnalyzerMap.put(MpegConstants.TS_STREAM_TYPE_LATM_AUDIO, new AdvancedAudioCodecAnalyzer());
			audioAnalyzerMap.put(MpegConstants.TS_STREAM_TYPE_AC3_AUDIO, new DolbyDigitalAnalzyer());
			audioAnalyzerMap.put(MpegConstants.TS_STREAM_TYPE_EAC3_AUDIO, new DolbyDigitalAnalzyer());
			
		}
		
		private StreamIdentifier videoStreamIdentifier=new StreamIdentifier(new StreamAnalyzerI[] {new MpegVideoAnalyzer(), new VC1VideoAnalyzer(), new AvcVideoAnalyzer()});
		
		
		public int addPacket(Packet packet) throws IOException {
			StreamDescriptor sd=avStreams[packet.getStreamId()];
			if (sd instanceof VideoStreamDescriptor) {
				StreamDescriptor vsd=videoStreamIdentifier.identifyStream(packet.getStreamId(), packet.getPayload());
				if (vsd instanceof VideoStreamDescriptor) {
					vsd.setLanguage(sd.getLanguage());
					avStreams[packet.getStreamId()]=vsd;
				}
			} if (sd instanceof AudioStreamDescriptor) {
				AudioStreamDescriptor asd=(AudioStreamDescriptor)audioAnalyzerMap.get(sd.getTsType()).analyze(packet.getPayload(), packet.getStreamId(), 0);
				if (asd != null) {
					asd.setLanguage(sd.getLanguage());
					avStreams[packet.getStreamId()]=asd;
				}
			}
			return MuxManager.CONTINUE;
		}
		public Map<Integer, StreamDescriptor> getStreamDescriptorMap() {
			HashMap<Integer, StreamDescriptor> streamDescriptorMap=new HashMap<Integer, StreamDescriptor>();
			for (int i=0;i<avStreams.length;i++) {
				if (avStreams[i] != null) {
					//Hack for Windows
					if (avStreams[i].getTsType()==MpegConstants.TS_STREAM_TYPE_EAC3_AUDIO) {
						avStreams[i].setTsType(MpegConstants.TS_STREAM_TYPE_AC3_AUDIO);
					}
					streamDescriptorMap.put(avStreams[i].getStreamId(), avStreams[i]);
				}
			}
			return streamDescriptorMap;
		}
		public int getPosition() {
			return POSITION_POST_BUFFER;
		}	
		
	}
	
	public Map<Integer, StreamDescriptor> getStreamDescriptorMap(JobStreamI stream, MuxManager muxManager) throws IOException {
		WtvStreamIdentifier identifier = new WtvStreamIdentifier();
		muxManager.add(identifier);
		stream.setStreamSize(Math.min(stream.getStreamSize(),100*1024*1024));
		muxManager.mux(stream);
		muxManager.remove(identifier);
		stream.close();
		return identifier.getStreamDescriptorMap();
	}
}
